@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2001, 2005, 2006, 2007,
-   2008 Free Software Foundation, Inc.
+   Copyright (C) 1999, 2001, 2005-2008, 2010-2012, 2014-2016 Free
+   Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,13 +13,12 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include "imap4d.h"
 
-struct imap_auth {
+struct imap_auth
+{
   char *name;
   imap4d_auth_handler_fp handler;
 };
@@ -36,10 +35,7 @@ comp (const void *item, const void *data)
 void
 auth_add (char *name, imap4d_auth_handler_fp handler)
 {
-  struct imap_auth *p = malloc (sizeof (*p));
-
-  if (!p)
-    imap4d_bye (ERR_NO_MEM);
+  struct imap_auth *p = mu_alloc (sizeof (*p));
 
   p->name = name;
   p->handler = handler;
@@ -47,6 +43,7 @@ auth_add (char *name, imap4d_auth_handler_fp handler)
     {
       mu_list_create (&imap_auth_list);
       mu_list_set_comparator (imap_auth_list, comp);
+      mu_list_set_destroy_item (imap_auth_list, mu_list_free_item);
     }
   mu_list_append (imap_auth_list, (void*)p);
 }
@@ -61,36 +58,30 @@ static int
 _auth_capa (void *item, void *usused)
 {
   struct imap_auth *p = item;
-  util_send(" AUTH=%s", p->name);
+  io_sendf (" AUTH=%s", p->name);
   return 0;
 }
 
-struct auth_data {
-  struct imap4d_command *command;
-  char *auth_type;
-  char *arg;
-  char *username;
-  int result;
-};
-
 static int
 _auth_try (void *item, void *data)
 {
   struct imap_auth *p = item;
-  struct auth_data *ap = data;
+  struct imap4d_auth *ap = data;
 
   if (strcmp (p->name, ap->auth_type) == 0)
     {
-      ap->result = p->handler (ap->command, ap->auth_type, &ap->username);
-      return 1;
+      int res = p->handler (ap);
+      if (res)
+	return res;
     }
   return 0;
 }
 
 void
-imap4d_auth_capability ()
+imap4d_auth_capability (struct imap4d_session *session)
 {
-  mu_list_do (imap_auth_list, _auth_capa, NULL);
+  if (session->tls_mode != tls_required)
+    mu_list_foreach (imap_auth_list, _auth_capa, NULL);
 }
 
 /*
@@ -100,40 +91,53 @@ imap4d_auth_capability ()
 */
 
 int
-imap4d_authenticate (struct imap4d_command *command, imap4d_tokbuf_t tok)
+imap4d_authenticate (struct imap4d_session *session,
+		     struct imap4d_command *command, imap4d_tokbuf_t tok)
 {
   char *auth_type;
-  struct auth_data adata;
-
+  struct imap4d_auth adata;
+  enum imap4d_auth_result res;
+  
   if (imap4d_tokbuf_argc (tok) != 3)
-    return util_finish (command, RESP_BAD, "Invalid arguments");
+    return io_completion_response (command, RESP_BAD, "Invalid arguments");
   
   auth_type = imap4d_tokbuf_getarg (tok, IMAP4_ARG_1);
 
-  if (tls_required)
-    return util_finish (command, RESP_NO,
-			"Command disabled: Use STARTTLS first");
+  if (session->tls_mode == tls_required)
+    return io_completion_response (command, RESP_NO,
+			           "Command disabled: Use STARTTLS first");
   
   adata.command = command;
   adata.auth_type = auth_type;
-  adata.arg = NULL;
   adata.username = NULL;
 
-  if (mu_list_do (imap_auth_list, _auth_try, &adata) == 0)
-    return util_finish (command, RESP_NO,
-			"Authentication mechanism not supported");
-  
-  if (adata.result == RESP_OK && adata.username)
+  res = mu_list_foreach (imap_auth_list, _auth_try, &adata);
+
+  switch (res)
     {
-      if (imap4d_session_setup (adata.username))
-	return util_finish (command, RESP_NO,
-			    "User name or passwd rejected");
-      else
-	return util_finish (command, RESP_OK,
-			    "%s authentication successful", auth_type);
+    case imap4d_auth_nosup:
+      return io_completion_response (command, RESP_NO,
+				     "Authentication mechanism not supported");
+    case imap4d_auth_ok:
+      return 0;
+
+    case imap4d_auth_resp:
+      if (adata.response == RESP_OK && adata.username)
+	{
+	  if (imap4d_session_setup (adata.username))
+	    return io_completion_response (command, RESP_NO,
+					   "User name or passwd rejected");
+	  else
+	    return io_completion_response (command, RESP_OK,
+					   "%s authentication successful",
+					   auth_type);
+	}
+      /* fall through */
+    case imap4d_auth_fail:
+      adata.response = RESP_NO;
+      break;
     }
-      
-  return util_finish (command, adata.result,
-		      "%s authentication failed", auth_type);
+  return io_completion_response (command, adata.response,
+				 "%s authentication failed", auth_type);
 }
 
