@@ -1,10 +1,11 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2005 Free Software Foundation, Inc.
+   Copyright (C) 1999-2001, 2005, 2007, 2010-2012, 2014-2016 Free
+   Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
+   version 3 of the License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -12,9 +13,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA */
+   Public License along with this library.  If not, see 
+   <http://www.gnu.org/licenses/>. */
 
 #ifndef _MAILUTILS_BODY_H
 #define _MAILUTILS_BODY_H
@@ -31,11 +31,13 @@ extern void * mu_body_get_owner   (mu_body_t);
 extern int mu_body_is_modified    (mu_body_t);
 extern int mu_body_clear_modified (mu_body_t);
 
-extern int mu_body_get_stream     (mu_body_t, mu_stream_t *);
+extern int mu_body_get_stream     (mu_body_t, mu_stream_t *) MU_DEPRECATED;
+extern int mu_body_get_streamref  (mu_body_t body, mu_stream_t *pstream);
 extern int mu_body_set_stream     (mu_body_t, mu_stream_t, void *owner);
-
-extern int mu_body_get_filename   (mu_body_t, char *, size_t, size_t *);
-
+extern int mu_body_set_get_stream (mu_body_t,
+				   int (*) (mu_body_t, mu_stream_t *),
+				   void *owner);
+  
 extern int mu_body_size           (mu_body_t, size_t *);
 extern int mu_body_set_size       (mu_body_t,
 				int (*_size) (mu_body_t, size_t*), void *owner);
