@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2005, 2007, 2008, 2009, 2010,
-   2011 Free Software Foundation, Inc.
+   Copyright (C) 1999-2002, 2005, 2007-2012, 2014-2016 Free Software
+   Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -24,198 +24,131 @@
 #include <assert.h>
 #include <sieve-priv.h>
 
-#define SIEVE_ARG(m,n,t) ((m)->prog[(m)->pc+(n)].t)
-#define SIEVE_ADJUST(m,n) (m)->pc+=(n)
+#define SIEVE_RT_ARG(m,n,t) ((m)->prog[(m)->pc+(n)].t)
+#define SIEVE_RT_ADJUST(m,n) (m)->pc+=(n)
 
+#define INSTR_DISASS(m) ((m)->state == mu_sieve_state_disass)
 #define INSTR_DEBUG(m) \
-  ((m)->debug_level & (MU_SIEVE_DEBUG_INSTR|MU_SIEVE_DEBUG_DISAS))
-#define INSTR_DISASS(m) ((m)->debug_level & MU_SIEVE_DEBUG_DISAS)
-
+  (INSTR_DISASS(m) || mu_debug_level_p (mu_sieve_debug_handle, MU_DEBUG_TRACE9)) 
+
 void
-_mu_sv_instr_nop (mu_sieve_machine_t mach)
+_mu_i_sv_instr_source (mu_sieve_machine_t mach)
 {
-  if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: NOP", (unsigned long) (mach->pc - 1));
-}
-
-void
-_mu_sv_instr_source (mu_sieve_machine_t mach)
-{
-  mach->locus.mu_file = (char*) SIEVE_ARG (mach, 0, string);
+  mach->locus.mu_file = mu_i_sv_id_str (mach, SIEVE_RT_ARG (mach, 0, pc));
   mu_stream_ioctl (mach->errstream, MU_IOCTL_LOGSTREAM,
                    MU_IOCTL_LOGSTREAM_SET_LOCUS,
 		   &mach->locus);
   if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: SOURCE %s",
-		    (unsigned long) (mach->pc - 1),
-		    mach->locus.mu_file);
-  SIEVE_ADJUST (mach, 1);
+    mu_i_sv_debug (mach, mach->pc - 1, "SOURCE %s", mach->locus.mu_file);
+  SIEVE_RT_ADJUST (mach, 1);
 }
 		 
 void
-_mu_sv_instr_line (mu_sieve_machine_t mach)
+_mu_i_sv_instr_line (mu_sieve_machine_t mach)
 {
-  mach->locus.mu_line = SIEVE_ARG (mach, 0, line);
+  mach->locus.mu_line = SIEVE_RT_ARG (mach, 0, line);
   mu_stream_ioctl (mach->errstream, MU_IOCTL_LOGSTREAM,
                    MU_IOCTL_LOGSTREAM_SET_LOCUS,
 		   &mach->locus);
   if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: LINE %u",
-		 (unsigned long) (mach->pc - 1),
-		 mach->locus.mu_line);
-  SIEVE_ADJUST (mach, 1);
+    mu_i_sv_debug (mach, mach->pc - 1, "LINE %u",
+		   mach->locus.mu_line);
+  SIEVE_RT_ADJUST (mach, 1);
 }
 		 
 static int
-instr_run (mu_sieve_machine_t mach)
+instr_run (mu_sieve_machine_t mach, char const *what)
 {
-  mu_sieve_handler_t han = SIEVE_ARG (mach, 0, handler);
-  mu_list_t arg_list = SIEVE_ARG (mach, 1, list);
-  mu_list_t tag_list = SIEVE_ARG (mach, 2, list);
   int rc = 0;
-  
-  SIEVE_ADJUST(mach, 4);
+  mu_sieve_handler_t han = SIEVE_RT_ARG (mach, 0, handler);
+  mach->argstart = SIEVE_RT_ARG (mach, 1, pc);
+  mach->argcount = SIEVE_RT_ARG (mach, 2, pc);
+  mach->tagcount = SIEVE_RT_ARG (mach, 3, pc);
+  mach->identifier = SIEVE_RT_ARG (mach, 4, string);
+  mach->comparator = SIEVE_RT_ARG (mach, 5, comp);
+				   
+  SIEVE_RT_ADJUST (mach, 6);
 
   if (INSTR_DEBUG (mach))
-    {
-      mu_stream_printf (mach->errstream,
-			"\033s<%d>Arguments: ", MU_LOG_DEBUG);
-      mu_sv_print_value_list (arg_list, mach->errstream);
-      mu_stream_printf (mach->errstream, "\n\033s<%d>Tags: ", MU_LOG_DEBUG);
-      mu_sv_print_tag_list (tag_list, mach->errstream);
-      mu_stream_printf (mach->errstream, "\n");
-    }
-
-  if (!INSTR_DISASS(mach))
-    rc = han (mach, arg_list, tag_list);
+    mu_i_sv_debug_command (mach, mach->pc - 7, what);
+  else
+    mu_i_sv_trace (mach, what);
+  
+  if (!INSTR_DISASS (mach))
+    rc = han (mach);
+  mach->argstart = 0;
+  mach->argcount = 0;
+  mach->tagcount = 0;
+  mach->identifier = NULL;
+  mach->comparator = NULL;
   return rc;
 }
 
 void
-_mu_sv_instr_action (mu_sieve_machine_t mach)
+_mu_i_sv_instr_action (mu_sieve_machine_t mach)
 {
-  mach->identifier = SIEVE_ARG (mach, 3, string);
-  if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: ACTION: %s",
-		    (unsigned long) (mach->pc - 1),
-		    mach->identifier);
   mach->action_count++;
-  instr_run (mach);
-  mach->identifier = NULL;
+  instr_run (mach, "ACTION");
 }
 
 void
-_mu_sv_instr_test (mu_sieve_machine_t mach)
+_mu_i_sv_instr_test (mu_sieve_machine_t mach)
 {
-  mach->identifier = SIEVE_ARG (mach, 3, string);
-  if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: TEST: %s",
-		    (unsigned long) (mach->pc - 1),
-		    mach->identifier);
-  mach->reg = instr_run (mach);
-  mach->identifier = NULL;
+  mach->reg = instr_run (mach, "TEST");
 }
 
 void
-_mu_sv_instr_push (mu_sieve_machine_t mach)
+_mu_i_sv_instr_not (mu_sieve_machine_t mach)
 {
   if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: PUSH", (unsigned long)(mach->pc - 1));
-      if (INSTR_DISASS (mach))
-	return;
-    }
-  
-  if (!mach->stack && mu_list_create (&mach->stack))
-    {
-      mu_sieve_error (mach, _("cannot create stack"));
-      mu_sieve_abort (mach);
-    }
-  mu_list_prepend (mach->stack, (void*) mach->reg);
-}
-
-void
-_mu_sv_instr_pop (mu_sieve_machine_t mach)
-{
-  if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: POP", (unsigned long)(mach->pc - 1));
-      if (INSTR_DISASS (mach))
-	return;
-    }
-
-  if (!mach->stack || mu_list_is_empty (mach->stack))
-    {
-      mu_sieve_error (mach, _("stack underflow"));
-      mu_sieve_abort (mach);
-    }
-  mu_list_get (mach->stack, 0, (void **)&mach->reg);
-  mu_list_remove (mach->stack, (void *)mach->reg);
-}
-
-void
-_mu_sv_instr_not (mu_sieve_machine_t mach)
-{
-  if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: NOT", (unsigned long)(mach->pc - 1));
-      if (INSTR_DISASS (mach))
-	return;
-    }
+    mu_i_sv_debug (mach, mach->pc - 1, "NOT");
+  if (INSTR_DISASS (mach))
+    return;
   mach->reg = !mach->reg;
 }
 
 void
-_mu_sv_instr_branch (mu_sieve_machine_t mach)
+_mu_i_sv_instr_branch (mu_sieve_machine_t mach)
 {
-  long num = SIEVE_ARG (mach, 0, number);
+  long num = SIEVE_RT_ARG (mach, 0, number);
 
-  SIEVE_ADJUST (mach, 1);
+  SIEVE_RT_ADJUST (mach, 1);
   if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: BRANCH %lu",
-		      (unsigned long)(mach->pc-2),
-		      (unsigned long)(mach->pc + num));
-      if (INSTR_DISASS (mach))
-	return;
-    }
+    mu_i_sv_debug (mach, mach->pc - 2, "BRANCH %lu",
+		   (unsigned long)(mach->pc + num));
+  if (INSTR_DISASS (mach))
+    return;
 
   mach->pc += num;
 }
 
 void
-_mu_sv_instr_brz (mu_sieve_machine_t mach)
+_mu_i_sv_instr_brz (mu_sieve_machine_t mach)
 {
-  long num = SIEVE_ARG (mach, 0, number);
-  SIEVE_ADJUST (mach, 1);
+  long num = SIEVE_RT_ARG (mach, 0, number);
+  SIEVE_RT_ADJUST (mach, 1);
 
   if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: BRZ %lu",
-		      (unsigned long)(mach->pc-2),
-		      (unsigned long)(mach->pc + num));
-      if (INSTR_DISASS (mach))
-	return;
-    }
+    mu_i_sv_debug (mach, mach->pc - 2, "BRZ %lu",
+		   (unsigned long)(mach->pc + num));
+  if (INSTR_DISASS (mach))
+    return;
   
   if (!mach->reg)
     mach->pc += num;
 }
   
 void
-_mu_sv_instr_brnz (mu_sieve_machine_t mach)
+_mu_i_sv_instr_brnz (mu_sieve_machine_t mach)
 {
-  long num = SIEVE_ARG (mach, 0, number);
-  SIEVE_ADJUST (mach, 1);
+  long num = SIEVE_RT_ARG (mach, 0, number);
+  SIEVE_RT_ADJUST (mach, 1);
 
   if (INSTR_DEBUG (mach))
-    {
-      mu_sieve_debug (mach, "%4lu: BRNZ %lu",
-		      (unsigned long)(mach->pc-2),
-		      (unsigned long)(mach->pc + num));
-      if (INSTR_DISASS (mach))
-	return;
-    }
+    mu_i_sv_debug (mach, mach->pc - 2, "BRNZ %lu",
+		   (unsigned long)(mach->pc + num));
+  if (INSTR_DISASS (mach))
+    return;
   
   if (mach->reg)
     mach->pc += num;
@@ -242,7 +175,7 @@ mu_sieve_get_data (mu_sieve_machine_t mach)
 int
 mu_sieve_get_locus (mu_sieve_machine_t mach, struct mu_locus *loc)
 {
-  if (mach->source_list)
+  if (mach->locus.mu_file)
     {
       *loc = mach->locus;
       return 0;
@@ -250,6 +183,12 @@ mu_sieve_get_locus (mu_sieve_machine_t mach, struct mu_locus *loc)
   return 1;
 }
 
+mu_mailbox_t
+mu_sieve_get_mailbox (mu_sieve_machine_t mach)
+{
+  return mach->mailbox;
+}
+
 mu_message_t
 mu_sieve_get_message (mu_sieve_machine_t mach)
 {
@@ -264,53 +203,75 @@ mu_sieve_get_message_num (mu_sieve_machine_t mach)
   return mach->msgno;
 }
 
-int
-mu_sieve_get_debug_level (mu_sieve_machine_t mach)
-{
-  return mach->debug_level;
-}
-
 const char *
 mu_sieve_get_identifier (mu_sieve_machine_t mach)
 {
   return mach->identifier;
 }
 
+void
+mu_sieve_get_argc (mu_sieve_machine_t mach, size_t *args, size_t *tags)
+{
+  if (args)
+    *args = mach->argcount;
+  if (tags)
+    *tags = mach->tagcount;
+}
+
 int
 mu_sieve_is_dry_run (mu_sieve_machine_t mach)
 {
-  return mach->debug_level & MU_SIEVE_DRY_RUN;
+  return mach->dry_run;
+}
+
+int
+mu_sieve_set_dry_run (mu_sieve_machine_t mach, int val)
+{
+  if (mach->state != mu_sieve_state_compiled)
+    return EINVAL; /* FIXME: another error code */
+  return mach->dry_run = val;
 }
 
 int
 sieve_run (mu_sieve_machine_t mach)
 {
-  if (setjmp (mach->errbuf))
-    return 1;
+  int rc;
 
-  mach->action_count = 0;
+  mu_sieve_stream_save (mach);
   
-  for (mach->pc = 1; mach->prog[mach->pc].handler; )
-    (*mach->prog[mach->pc++].instr) (mach);
+  rc = setjmp (mach->errbuf);
+  if (rc == 0)
+    {
+      mach->action_count = 0;
 
-  if (mach->action_count == 0)
-    mu_sieve_log_action (mach, "IMPLICIT KEEP", NULL);
+      if (mu_sieve_has_variables (mach))
+	mu_assoc_clear (mach->vartab);
+      
+      for (mach->pc = 1; mach->prog[mach->pc].handler; )
+	(*mach->prog[mach->pc++].instr) (mach);
+
+      if (mach->action_count == 0)
+	mu_sieve_log_action (mach, "IMPLICIT KEEP", NULL);
   
-  if (INSTR_DEBUG (mach))
-    mu_sieve_debug (mach, "%4lu: STOP", (unsigned long) mach->pc);
+      if (INSTR_DEBUG (mach))
+	mu_i_sv_debug (mach, mach->pc, "STOP");
+    }
   
-  return 0;
+  mu_sieve_stream_restore (mach);
+  
+  return rc;
 }
 
 int
 mu_sieve_disass (mu_sieve_machine_t mach)
 {
-  int level = mach->debug_level;
   int rc;
 
-  mach->debug_level = MU_SIEVE_DEBUG_INSTR | MU_SIEVE_DEBUG_DISAS;
+  if (mach->state != mu_sieve_state_compiled)
+    return EINVAL; /* FIXME: Error code */
+  mach->state = mu_sieve_state_disass;
   rc = sieve_run (mach);
-  mach->debug_level = level;
+  mach->state = mu_sieve_state_compiled;
   return rc;
 }
   
@@ -340,6 +301,10 @@ mu_sieve_mailbox (mu_sieve_machine_t mach, mu_mailbox_t mbox)
   if (!mach || !mbox)
     return EINVAL;
 
+  if (mach->state != mu_sieve_state_compiled)
+    return EINVAL; /* FIXME: Error code */
+  mach->state = mu_sieve_state_running;
+  
   mu_observer_create (&observer, mach);
   mu_observer_set_action (observer, _sieve_action, mach);
   mu_mailbox_get_observable (mbox, &observable);
@@ -354,6 +319,7 @@ mu_sieve_mailbox (mu_sieve_machine_t mach, mu_mailbox_t mbox)
   mu_observable_detach (observable, observer);
   mu_observer_destroy (&observer, mach);
 
+  mach->state = mu_sieve_state_compiled;
   mach->mailbox = NULL;
   
   return rc;
@@ -367,10 +333,15 @@ mu_sieve_message (mu_sieve_machine_t mach, mu_message_t msg)
   if (!mach || !msg)
     return EINVAL;
 
+  if (mach->state != mu_sieve_state_compiled)
+    return EINVAL; /* FIXME: Error code */
+  mach->state = mu_sieve_state_running;
+
   mach->msgno = 1;
   mach->msg = msg;
   mach->mailbox = NULL;
   rc = sieve_run (mach);
+  mach->state = mu_sieve_state_compiled;
   mach->msg = NULL;
   
   return rc;
