@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2003, 
-   2005, 2006, 2007, 2008 Free Software Foundation, Inc.
+   Copyright (C) 1999-2003, 2005-2012, 2014-2016 Free Software
+   Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,9 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 /* MH folder command */
 
@@ -30,86 +28,16 @@
 
 #include <dirent.h>
 
-#define obstack_chunk_alloc malloc
-#define obstack_chunk_free free
-#include <obstack.h>
+static char prog_doc[] = N_("set or list current folder or message");
+static char args_doc[] = N_("[ACTION] [MSG]");
 
-const char *program_version = "folder (" PACKAGE_STRING ")";
-static char doc[] = N_("GNU MH folder")"\v"
-N_("Use -help to obtain the list of traditional MH options.");
-static char args_doc[] = N_("[action] [msg]");
+typedef int (*folder_action) (void);
 
-static struct argp_option options[] = {
-  {N_("Actions are:"), 0, 0, OPTION_DOC, NULL, 0 },
-  {"print", ARG_PRINT, NULL, 0,
-   N_("List the folders (default)"), 1 },
-  {"list",  ARG_LIST,  NULL, 0,
-   N_("List the contents of the folder stack"), 1},
-  {"pack",  ARG_PACK,  N_("NUMBER"), OPTION_ARG_OPTIONAL,
-   N_("Remove holes in message numbering. Begin numbering from NUMBER (default: first message number)"), 1},
-  {"push",  ARG_PUSH,  N_("FOLDER"), OPTION_ARG_OPTIONAL,
-    N_("Push the folder on the folder stack. If FOLDER is specified, it is pushed. "
-       "Otherwise, if a folder is given in the command line (via + or --folder), "
-       "it is pushed on stack. Otherwise, the current folder and the top of the folder "
-       "stack are exchanged"), 1},
-  {"pop",   ARG_POP,    NULL, 0,
-   N_("Pop the folder off the folder stack"), 1},
-  
-  {N_("Options are:"), 0, 0, OPTION_DOC, NULL, 2 },
-  
-  {"folder", ARG_FOLDER, N_("FOLDER"), 0,
-   N_("Specify folder to operate upon"), 3},
-  {"all",    ARG_ALL,    NULL, 0,
-   N_("List all folders"), 3},
-  {"create", ARG_CREATE, N_("BOOL"), OPTION_ARG_OPTIONAL, 
-    N_("Create non-existing folders"), 3},
-  {"nocreate", ARG_NOCREATE, NULL, OPTION_HIDDEN, ""},
-  {"fast",   ARG_FAST, N_("BOOL"), OPTION_ARG_OPTIONAL, 
-    N_("List only the folder names"), 3},
-  {"nofast", ARG_NOFAST, NULL, OPTION_HIDDEN, ""},
-  {"header", ARG_HEADER, N_("BOOL"), OPTION_ARG_OPTIONAL, 
-    N_("Print the header line"), 3},
-  {"noheader", ARG_NOHEADER, NULL, OPTION_HIDDEN, ""},
-  {"recurse",ARG_RECURSIVE, N_("BOOL"), OPTION_ARG_OPTIONAL,
-    N_("Scan folders recursively"), 3},
-  {"norecurse", ARG_NORECURSIVE, NULL, OPTION_HIDDEN, ""},
-  {"total",  ARG_TOTAL, N_("BOOL"), OPTION_ARG_OPTIONAL, 
-    N_("Output the total statistics"), 3},
-  {"nototal", ARG_NOTOTAL, NULL, OPTION_HIDDEN, ""},
-  {"verbose", ARG_VERBOSE, NULL, 0,
-   N_("Verbosely list actions taken"), 3},
-  {"dry-run", ARG_DRY_RUN, NULL, 0,
-   N_("Do nothing, print what would be done (with --pack)"), 3},
-   
-  {"license", ARG_LICENSE, 0,      0,
-   N_("Display software license"), -1},
-
-  {NULL},
-};
-
-/* Traditional MH options */
-struct mh_option mh_option[] = {
-  {"print",   2, 0, NULL },
-  {"list",    1, 0, NULL },
-  {"push",    2, 0, NULL },
-  {"pop",     2, 0, NULL },
-  {"all",     1, 0, NULL },
-  {"pack",    2, 0, NULL },
-  {"create",  1, MH_OPT_BOOL, NULL},
-  {"fast",    1, MH_OPT_BOOL, NULL},
-  {"header",  1, MH_OPT_BOOL, NULL},
-  {"recurse", 1, MH_OPT_BOOL, NULL},
-  {"total",   1, MH_OPT_BOOL, NULL},
-  {NULL},
-};
-
-typedef int (*folder_action) ();
-
-static int action_print ();
-static int action_list ();
-static int action_push ();
-static int action_pop ();
-static int action_pack ();
+static int action_print (void);
+static int action_list (void);
+static int action_push (void);
+static int action_pop (void);
+static int action_pack (void);
 static folder_action action = action_print;
 
 int show_all = 0; /* List all folders. Raised by --all switch */
@@ -131,114 +59,105 @@ const char *push_folder; /* Folder name to push on stack */
 const char *mh_seq_name; /* Name of the mh sequence file (defaults to
 			    .mh_sequences) */
 int has_folder;    /* Folder has been explicitely given */
+int recurse_option = 0;
 size_t max_depth = 1;  /* Maximum recursion depth (0 means infinity) */ 
 
 #define OPTION_IS_SET(opt) ((opt) == -1 ? show_all : opt)
 
-static int
-opt_handler (int key, char *arg, void *unused, struct argp_state *state)
+void
+set_action (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
 {
-  switch (key)
+  if (strcmp (opt->opt_long, "print") == 0)
+    action = action_print;
+  else if (strcmp (opt->opt_long, "pack") == 0)
     {
-    case ARG_DRY_RUN:
-      dry_run++;
-      break;
-	
-    case ARG_PACK:
       action = action_pack;
       if (arg)
 	{
-	  char *p;
-	  pack_start = strtoul (arg, &p, 10);
-	  if (*p)
-	    argp_error (state, _("Invalid number"));
+	  if (mu_str_to_c (arg, mu_c_size, &pack_start, NULL))
+	    {
+	      mu_parseopt_error (po, _("%s: invalid number"), arg);
+	      exit (po->po_exit_error);
+	    }
 	}
-      break;
-      
-    case ARG_PRINT:
-      action = action_print;
-      break;
-      
-    case ARG_LIST:
-      action = action_list;
-      break;
-
-    case ARG_PUSH:
+    }
+  else if (strcmp (opt->opt_long, "list") == 0)
+    action = action_list;
+  else if (strcmp (opt->opt_long, "push") == 0)
+    {
       action = action_push;
       if (arg)
 	{
 	  push_folder = mh_current_folder ();
 	  mh_set_current_folder (arg);
 	}
-      break;
-      
-    case ARG_POP:
-      action = action_pop;
-      break;
-      
-    case ARG_ALL:
-      show_all = 1;
-      break;
-
-    case ARG_CREATE:
-      create_flag = is_true (arg);
-      break;
-
-    case ARG_NOCREATE:
-      create_flag = 0;
-      
-    case ARG_FAST:
-      fast_mode = is_true (arg);
-      break;
-
-    case ARG_NOFAST:
-      fast_mode = 0;
-      break;
-      
-    case ARG_HEADER:
-      print_header = is_true (arg);
-      break;
-
-    case ARG_NOHEADER:
-      print_header = 0;
-      break;
-      
-    case ARG_RECURSIVE:
-      max_depth = is_true (arg) ? 0 : 1;
-      break;
-
-    case ARG_NORECURSIVE:
-      max_depth = 0;
-      break;
-      
-    case ARG_TOTAL:
-      print_total = is_true (arg);
-      break;
-
-    case ARG_NOTOTAL:
-      print_total = 0;
-      break;
-      
-    case ARG_FOLDER:
-      has_folder = 1;
-      push_folder = mh_current_folder ();
-      mh_set_current_folder (arg);
-      break;
-      
-    case ARG_LICENSE:
-      mh_license (argp_program_version);
-      break;
-
-    case ARG_VERBOSE:
-      verbose++;
-      break;
-      
-    default:
-      return 1;
     }
-  return 0;
+  else if (strcmp (opt->opt_long, "pop") == 0)
+    action = action_pop;
+  else
+    abort ();
 }
 
+void
+set_folder (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
+{
+  has_folder = 1;
+  push_folder = mh_current_folder ();
+  mh_set_current_folder (arg);
+}
+
+static struct mu_option options[] = {
+  MU_OPTION_GROUP (N_("Actions are:")),
+  { "print", 0, NULL, MU_OPTION_DEFAULT,
+    N_("list the folders (default)"),
+    mu_c_string, NULL, set_action },
+  { "list",   0, NULL, MU_OPTION_DEFAULT,
+    N_("list the contents of the folder stack"),
+    mu_c_string, NULL, set_action },
+  { "pack",   0, N_("NUMBER"), MU_OPTION_ARG_OPTIONAL,
+    N_("remove holes in message numbering, begin numbering from NUMBER (default: first message number)"),
+    mu_c_string, NULL, set_action },
+  { "push",   0, N_("FOLDER"), MU_OPTION_ARG_OPTIONAL,
+    N_("push the folder on the folder stack. If FOLDER is specified, it is pushed. "
+       "Otherwise, if a folder is given in the command line (via + or --folder), "
+       "it is pushed on stack. Otherwise, the current folder and the top of the folder "
+       "stack are exchanged"),
+    mu_c_string, NULL, set_action },
+  { "pop",    0, NULL, MU_OPTION_DEFAULT,
+    N_("pop the folder off the folder stack"), 
+    mu_c_string, NULL, set_action },
+  
+  MU_OPTION_GROUP (N_("Options are:")),
+  
+  { "folder", 0, N_("FOLDER"), MU_OPTION_DEFAULT,
+    N_("specify folder to operate upon"),
+    mu_c_string, NULL, set_folder },
+  { "all",    0,    NULL, MU_OPTION_DEFAULT,
+    N_("list all folders"),
+    mu_c_bool, &show_all },
+  { "create", 0, NULL, MU_OPTION_DEFAULT, 
+    N_("create non-existing folders"),
+    mu_c_bool, &create_flag },
+  { "fast",   0, NULL, MU_OPTION_DEFAULT, 
+    N_("list only the folder names"),
+    mu_c_bool, &fast_mode },
+  { "header", 0, NULL, MU_OPTION_DEFAULT, 
+    N_("print the header line"),
+    mu_c_bool, &print_header },
+  { "recurse",0, NULL, MU_OPTION_DEFAULT,
+    N_("scan folders recursively"),
+    mu_c_bool, &recurse_option },
+  { "total",  0, NULL, MU_OPTION_DEFAULT, 
+    N_("output the total statistics"),
+    mu_c_bool, &print_total },
+  { "verbose", 0, NULL, MU_OPTION_DEFAULT,
+    N_("verbosely list actions taken"),
+    mu_c_bool, &verbose },
+  { "dry-run", 0, NULL, MU_OPTION_DEFAULT,
+    N_("do nothing, print what would be done (with --pack)"),
+    mu_c_bool, &dry_run },
+  MU_OPTION_END
+};
 
 /* ************************************************************* */
 /* Printing */
@@ -253,21 +172,19 @@ struct folder_info
   size_t others;           /* Number of non-message files */ 
 };
 
-struct obstack folder_info_stack; /* Memory storage for folder infp */
-struct folder_info *folder_info;  /* Finalized array of information
-				     structures */
-size_t folder_info_count;         /* Number of the entries in the array */
+mu_list_t folder_info_list; /* Memory storage for folder info */
 
 size_t message_count;             /* Total number of messages */
 
 int name_prefix_len;              /* Length of the mu_path_folder_dir */
 
 void
-install_folder_info (const char *name, struct folder_info *info)
+install_folder_info (const char *name, struct folder_info const *info)
 {
-  info->name = strdup (name) + name_prefix_len;
-  obstack_grow (&folder_info_stack, info, sizeof (*info));
-  folder_info_count++;
+  struct folder_info *new_info = mu_alloc (sizeof (*new_info));
+  *new_info = *info;
+  new_info->name = mu_strdup (new_info->name + name_prefix_len);
+  mu_list_append (folder_info_list, new_info);
   message_count += info->message_count;
 }
 
@@ -282,20 +199,15 @@ static void
 read_seq_file (struct folder_info *info, const char *prefix, const char *name)
 {
   char *pname = NULL;
-  mh_context_t *ctx;
+  mu_property_t prop;
   const char *p;
   
-  asprintf (&pname, "%s/%s", prefix, name);
-  if (!pname)
-    abort ();
-  ctx = mh_context_create (pname, 1);
-  mh_context_read (ctx);
+  pname = mh_safe_make_file_name (prefix, name);
+  prop = mh_read_property_file (pname, 1);
   
-  p = mh_context_get_value (ctx, "cur", NULL);
-  if (p)
+  if (mu_property_sget_value (prop, "cur", &p) == 0)
     info->cur = strtoul (p, NULL, 0);
-  free (pname);
-  free (ctx);
+  mu_property_destroy (&prop);
 }
 
 static void
@@ -327,7 +239,7 @@ _scan (const char *name, size_t depth)
 
   if (!dir)
     {
-      mu_error (_("Cannot scan folder %s: %s"), name, strerror (errno));
+      mu_error (_("cannot scan folder %s: %s"), name, strerror (errno));
       return;
     }
 
@@ -336,7 +248,7 @@ _scan (const char *name, size_t depth)
       if (fast_mode && depth > 0)
 	{
 	  memset (&info, 0, sizeof (info));
-	  info.name = strdup (name);
+	  info.name = (char*) name;
 	  install_folder_info (name, &info);
 	  closedir (dir);
 	  return;
@@ -350,7 +262,7 @@ _scan (const char *name, size_t depth)
     }
   
   memset (&info, 0, sizeof (info));
-  info.name = strdup (name);
+  info.name = mu_strdup (name);
   while ((entry = readdir (dir)))
     {
       if (entry->d_name[0] == '.')
@@ -360,9 +272,9 @@ _scan (const char *name, size_t depth)
 	}
       else if (entry->d_name[0] != ',')
 	{
-	  asprintf (&p, "%s/%s", name, entry->d_name);
+	  p = mh_safe_make_file_name (name, entry->d_name);
 	  if (stat (p, &st) < 0)
-	    mu_error (_("Cannot stat %s: %s"), p, strerror (errno));
+	    mu_diag_funcall (MU_DIAG_ERROR, "stat", p, errno);
 	  else if (S_ISDIR (st.st_mode))
 	    {
 	      info.others++;
@@ -388,7 +300,7 @@ _scan (const char *name, size_t depth)
   
   if (info.cur)
     {
-      asprintf (&p, "%s/%s", name, mu_umaxtostr (0, info.cur));
+      p = mh_safe_make_file_name (name, mu_umaxtostr (0, info.cur));
       if (stat (p, &st) < 0 || !S_ISREG (st.st_mode))
 	info.cur = 0;
       free (p);
@@ -397,64 +309,73 @@ _scan (const char *name, size_t depth)
   if (depth > 0)
     install_folder_info (name, &info);
 }
-    
-static void
-print_all ()
-{
-  struct folder_info *info, *end = folder_info + folder_info_count;
 
-  for (info = folder_info; info < end; info++)
+static int
+_folder_info_printer (void *item, void *data)
+{
+  struct folder_info *info = item;
+  int len = strlen (info->name);
+
+  if (len < 22)
+    printf ("%22.22s", info->name);
+  else
+    printf ("%s", info->name);
+  
+  if (strcmp (info->name, mh_current_folder ()) == 0)
+    printf ("+");
+  else
+    printf (" ");
+  
+  if (info->message_count)
     {
-      int len = strlen (info->name);
-      if (len < 22)
-	printf ("%22.22s", info->name);
-      else
-	printf ("%s", info->name);
-      
-      if (strcmp (info->name, mh_current_folder ()) == 0)
-	printf ("+");
-      else
-	printf (" ");
-      
-      if (info->message_count)
-	{
-	  printf (ngettext(" has %4lu message  (%4lu-%4lu)",
-			   " has %4lu messages (%4lu-%4lu)",
-			   info->message_count),
-		  (unsigned long) info->message_count,
-		  (unsigned long) info->min,
-		  (unsigned long) info->max);
-	  if (info->cur)
-	    printf ("; cur=%4lu", (unsigned long) info->cur);
-	}
-      else
-	{
-	  printf (_(" has no messages"));
-	}
-      
-      if (info->others)
-	{
-	  if (!info->cur)
-	    printf (";           ");
-	  else
-	    printf ("; ");
-	  printf (_("(others)"));
-	}
-      printf (".\n");
+      printf (ngettext(" has %4lu message  (%4lu-%4lu)",
+		       " has %4lu messages (%4lu-%4lu)",
+		       info->message_count),
+	      (unsigned long) info->message_count,
+	      (unsigned long) info->min,
+	      (unsigned long) info->max);
+      if (info->cur)
+	printf ("; cur=%4lu", (unsigned long) info->cur);
     }
-}
-
-static void
-print_fast ()
-{
-  struct folder_info *info, *end = folder_info + folder_info_count;
-
-  for (info = folder_info; info < end; info++)
-    printf ("%s\n", info->name);
+  else
+    {
+      printf (_(" has no messages"));
+    }
+  
+  if (info->others)
+    {
+      if (!info->cur)
+	printf (";           ");
+      else
+	printf ("; ");
+      printf (_("(others)"));
+    }
+  printf (".\n");
+  return 0;
 }
 
 static int
-action_print ()
+_folder_name_printer (void *item, void *data)
+{
+  struct folder_info *info = item;
+  printf ("%s\n", info->name);
+  return 0;
+}
+
+static void
+print_all (void)
+{
+  mu_list_foreach (folder_info_list, _folder_info_printer, NULL);
+}
+
+static void
+print_fast (void)
+{
+  mu_list_foreach (folder_info_list, _folder_name_printer, NULL);
+}
+
+static int
+action_print (void)
 {
   const char *folder_dir = mu_folder_directory ();
   mh_seq_name = mh_global_profile_get ("mh-sequences", MH_SEQUENCES_FILE);
@@ -464,7 +385,7 @@ action_print ()
     name_prefix_len++;
   name_prefix_len++;  /* skip past the slash */
 
-  obstack_init (&folder_info_stack);
+  mu_list_create (&folder_info_list);
 
   if (show_all)
     {
@@ -472,14 +393,12 @@ action_print ()
     }
   else
     {
-      char *p = mh_expand_name (NULL, mh_current_folder (), 0);
+      char *p = mh_expand_name (NULL, mh_current_folder (), NAME_ANY);
       _scan (p, 1);
       free (p);
     }
   
-  folder_info = obstack_finish (&folder_info_stack);
-  qsort (folder_info, folder_info_count, sizeof (folder_info[0]),
-	 folder_info_comp);
+  mu_list_sort (folder_info_list, folder_info_comp);
 
   if (fast_mode)
     print_fast ();
@@ -492,13 +411,16 @@ action_print ()
 
       if (OPTION_IS_SET (print_total))
 	{
+	  size_t folder_count;
+
+	  mu_list_count (folder_info_list, &folder_count);
 	  printf ("\n%24.24s=", _("TOTAL"));
 	  printf (ngettext ("%4lu message  ", "%4lu messages ",
 			    message_count),
 		  (unsigned long) message_count);
 	  printf (ngettext ("in %4lu folder", "in %4lu folders",
-			    folder_info_count),
-		  (unsigned long) folder_info_count);
+			    folder_count),
+		  (unsigned long) folder_count);
 	  printf ("\n");
 	}
     }
@@ -513,12 +435,12 @@ action_print ()
 /* Listing */
 
 static int
-action_list ()
+action_list (void)
 {
   const char *stack = mh_global_context_get ("Folder-Stack", NULL);
 
   printf ("%s", mh_current_folder ());
-  if (stack)
+  if (stack && stack[0])
     printf (" %s", stack);
   printf ("\n");
   return 0;
@@ -529,19 +451,26 @@ action_list ()
 /* Push & pop */
 
 static void
-get_stack (int *pc, char ***pv)
+get_stack (size_t *pc, char ***pv)
 {
+  struct mu_wordsplit ws;
   const char *stack = mh_global_context_get ("Folder-Stack", NULL);
   if (!stack)
     {
       *pc = 0;
       *pv = NULL;
     }
-  else if (mu_argcv_get (stack, NULL, "#", pc, pv))
+  else if (mu_wordsplit (stack, &ws, MU_WRDSF_DEFFLAGS))
     {
-      mu_error (_("Cannot split line %s"), stack);
+      mu_error (_("cannot split line `%s': %s"), stack,
+		mu_wordsplit_strerror (&ws));
       exit (1);
     }
+  else
+    {
+      mu_wordsplit_get_words (&ws, pc, pv);
+      mu_wordsplit_free (&ws);
+    }
 }
 
 static void
@@ -560,32 +489,32 @@ set_stack (int c, char **v)
 }
 
 static void
-push_val (int *pc, char ***pv, const char *val)
+push_val (size_t *pc, char ***pv, const char *val)
 {
-  int c = *pc;
+  size_t c = *pc;
   char **v = *pv;
 
   c++;
   if (c == 1)
     {
-      v = xcalloc (c + 1, sizeof (*v));
+      v = mu_calloc (c + 1, sizeof (*v));
     }
   else
     {
-      v = xrealloc (v, (c + 1) * sizeof (*v));
+      v = mu_realloc (v, (c + 1) * sizeof (*v));
       memmove (&v[1], &v[0], c * sizeof (*v));
     }
-  v[0] = xstrdup (val);
+  v[0] = mu_strdup (val);
 
   *pv = v;
   *pc = c;
 }
 
 static char *
-pop_val (int *pc, char ***pv)
+pop_val (size_t *pc, char ***pv)
 {
   char *val;
-  int c;
+  size_t c;
   char **v;
   
   if (*pc == 0)
@@ -602,9 +531,9 @@ pop_val (int *pc, char ***pv)
 }
   
 static int
-action_push ()
+action_push (void)
 {
-  int c;
+  size_t c;
   char **v;
 
   get_stack (&c, &v);
@@ -614,7 +543,7 @@ action_push ()
   else 
     {
       char *t = v[0];
-      v[0] = xstrdup (mh_current_folder ());
+      v[0] = mu_strdup (mh_current_folder ());
       mh_set_current_folder (t);
       free (t);
     }
@@ -627,9 +556,9 @@ action_push ()
 }
 
 static int
-action_pop ()
+action_pop (void)
 {
-  int c;
+  size_t c;
   char **v;
 
   get_stack (&c, &v);
@@ -721,9 +650,9 @@ roll_back (const char *folder_name, struct pack_tab *pack_tab, size_t i)
   if (i == 0)
     return;
   
-  start = i - 1;
-  mu_error (_("Rolling back changes..."));
-  while (--i >= 0)
+  start = --i;
+  mu_error (_("rolling back changes..."));
+  do
     if (pack_rename (pack_tab + i, 1))
       {
 	mu_error (_("CRITICAL ERROR: Folder `%s' left in an inconsistent state, because an error\n"
@@ -737,11 +666,13 @@ roll_back (const char *folder_name, struct pack_tab *pack_tab, size_t i)
 	mu_error (_("You will have to fix it manually."));
 	exit (1);
       }
-  mu_error (_("Folder `%s' restored successfully"), folder_name);
+  while (i-- > 0);
+  mu_error (_("folder `%s' restored successfully"), folder_name);
 }
 
 struct fixup_data
 {
+  mu_mailbox_t mbox;
   const char *folder_dir;
   struct pack_tab *pack_tab;
   size_t count;
@@ -770,45 +701,63 @@ pack_xlate (struct pack_tab *pack_tab, size_t count, size_t n)
   return p ? p->new : 0;
 }
 
-static int
+static void
 _fixup (const char *name, const char *value, struct fixup_data *fd, int flags)
 {
-  int i, j, argc;
-  char **argv;
-  mh_msgset_t msgset;
+  size_t i;
+  int rc;
+  struct mu_wordsplit ws;
+  mu_msgset_t msgset;
 
   if (verbose)
     fprintf (stderr, "Sequence `%s'...\n", name);
-  
-  if (mu_argcv_get (value, "", NULL, &argc, &argv))
-    return 0;
 
-  msgset.list = xcalloc (argc, sizeof msgset.list[0]);
-  for (i = j = 0; i < argc; i++)
+  if (mu_wordsplit (value, &ws, MU_WRDSF_DEFFLAGS))
+    {
+      mu_error (_("cannot split line `%s': %s"), value,
+		mu_wordsplit_strerror (&ws));
+      return;
+    }
+
+  rc = mu_msgset_create (&msgset, fd->mbox, MU_MSGSET_UID);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_msgset_create", NULL, rc);
+      exit (1);
+    }
+  
+  for (i = 0; i < ws.ws_wordc; i++)
     {
       size_t n = pack_xlate (fd->pack_tab, fd->count,
-			     strtoul (argv[i], NULL, 0));
+			     strtoul (ws.ws_wordv[i], NULL, 0));
       if (n)
-	msgset.list[j++] = n;
+	{
+	  rc = mu_msgset_add_range (msgset, n, n, MU_MSGSET_UID);
+	  if (rc)
+	    {
+	      mu_diag_funcall (MU_DIAG_ERROR, "mu_msgset_add_range", NULL, rc);
+	      exit (1);
+	    }
+	}
     }
-  msgset.count = j;
 
-  mh_seq_add (name, &msgset, flags | SEQ_ZERO);
-  free (msgset.list);
+  mu_wordsplit_free (&ws);
+  
+  mh_seq_add (fd->mbox, name, msgset, flags | SEQ_ZERO);
+  mu_msgset_free (msgset);
 
   if (verbose)
     {
-      const char *p = mh_seq_read (name, flags);
+      const char *p = mh_seq_read (fd->mbox, name, flags);
       fprintf (stderr, "Sequence %s: %s\n", name, p);
     }
-  
-  return 0;
 }
 
 static int
 fixup_global (const char *name, const char *value, void *data)
 {
-  return _fixup (name, value, data, 0);
+  _fixup (name, value, data, 0);
+  return 0;
 }
 
 static int
@@ -823,21 +772,21 @@ fixup_private (const char *name, const char *value, void *data)
   nlen = strlen (name) - strlen (fd->folder_dir);
   if (nlen > 0 && strcmp (name + nlen, fd->folder_dir) == 0)
     {
-      int rc;
-      char *s = xmalloc (nlen);
+      char *s = mu_alloc (nlen);
       memcpy (s, name, nlen - 1);
       s[nlen-1] = 0;
-      rc = _fixup (s, value, fd, SEQ_PRIVATE);
+      _fixup (s, value, fd, SEQ_PRIVATE);
       free (s);
     }
   return 0;
 }
 
 int
-action_pack ()
+action_pack (void)
 {
-  const char *folder_dir = mh_expand_name (NULL, mh_current_folder (), 0);
-  mu_mailbox_t mbox = mh_open_folder (mh_current_folder (), 0);
+  const char *folder_dir = mh_expand_name (NULL, mh_current_folder (), 
+                                           NAME_ANY);
+  mu_mailbox_t mbox = mh_open_folder (mh_current_folder (), MU_STREAM_RDWR);
   struct pack_tab *pack_tab;
   size_t i, count, start;
   int status;
@@ -846,11 +795,11 @@ action_pack ()
   /* Allocate pack table */
   if (mu_mailbox_messages_count (mbox, &count))
     {
-      mu_error (_("Cannot read input mailbox: %s"), mu_strerror (errno));
+      mu_error (_("cannot read input mailbox: %s"), mu_strerror (errno));
       return 1;
     }
-  pack_tab = xcalloc (count, sizeof pack_tab[0]); /* Never freed. No use to
-						     try to. */
+  pack_tab = mu_calloc (count, sizeof pack_tab[0]); /* Never freed. No use to
+		 				       try to. */
 
   /* Populate it with message numbers */
   if (verbose)
@@ -862,7 +811,8 @@ action_pack ()
       status = mu_mailbox_get_message (mbox, i + 1, &msg);
       if (status)
 	{
-	  mu_error (_("%d: cannot get message: %s"), i, mu_strerror (status));
+	  mu_error (_("%lu: cannot get message: %s"),
+		    (unsigned long) i, mu_strerror (status));
 	  return 1;
 	}
       mh_message_number (msg, &pack_tab[i].orig);
@@ -912,18 +862,23 @@ action_pack ()
     fprintf (stderr, _("Finished packing messages.\n"));
 
   /* Fix-up sequences */
-  fd.folder_dir = folder_dir;
-  fd.pack_tab = pack_tab;
-  fd.count = count;
-  if (verbose)
-    fprintf (stderr, _("Fixing global sequences\n"));
-  mh_global_sequences_iterate (fixup_global, &fd);
-  if (verbose)
-    fprintf (stderr, _("Fixing private sequences\n"));
-  mh_global_context_iterate (fixup_private, &fd);
-
   if (!dry_run)
-    mh_global_save_state ();
+    {
+      mbox = mh_open_folder (mh_current_folder (), MU_STREAM_RDWR);
+      fd.mbox = mbox;
+      fd.folder_dir = folder_dir;
+      fd.pack_tab = pack_tab;
+      fd.count = count;
+      if (verbose)
+	fprintf (stderr, _("Fixing global sequences\n"));
+      mh_global_sequences_iterate (mbox, fixup_global, &fd);
+      if (verbose)
+	fprintf (stderr, _("Fixing private sequences\n"));
+      mh_global_context_iterate (fixup_private, &fd);
+      mu_mailbox_close (mbox);
+      mu_mailbox_destroy (&mbox);
+      mh_global_save_state ();
+    }
   
   return 0;
 }
@@ -932,14 +887,11 @@ int
 main (int argc, char **argv)
 {
   int index = 0;
-  mh_msgset_t msgset;
+  mu_msgset_t msgset;
 
-  /* Native Language Support */
-  MU_APP_INIT_NLS ();
-
-  mh_argp_init (program_version);
-  mh_argp_parse (&argc, &argv, 0, options, mh_option, args_doc, doc,
-		 opt_handler, NULL, &index);
+  mh_getopt (&argc, &argv, options, 0, args_doc, prog_doc, NULL);
+  if (recurse_option)
+    max_depth = 0;
 
   /* If  folder  is invoked by a name ending with "s" (e.g.,  folders),
      `-all'  is  assumed */
@@ -959,16 +911,18 @@ main (int argc, char **argv)
     
   if (argc - index == 1)
     {
-      mu_mailbox_t mbox = mh_open_folder (mh_current_folder (), 0);
-      mh_msgset_parse (mbox, &msgset, argc - index, argv + index, "cur");
-      mh_msgset_current (mbox, &msgset, 0);
+      mu_mailbox_t mbox = mh_open_folder (mh_current_folder (),
+					  MU_STREAM_RDWR);
+      mh_msgset_parse (&msgset, mbox, argc - index, argv + index, "cur");
+      mh_msgset_first_current (mbox, msgset);
+      mu_msgset_free (msgset);
       mh_global_save_state ();
       mu_mailbox_close (mbox);
       mu_mailbox_destroy (&mbox);
     }
   else if (argc - index > 1)
     {
-      mu_error (_("Too many arguments"));
+      mu_error (_("too many arguments"));
       exit (1);
     }
   
