@@ -1,5 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2005, 2006, 2007 Free Software Foundation, Inc.
+   Copyright (C) 2003, 2005-2007, 2009-2012, 2014-2016 Free Software
+   Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,9 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with GNU Mailutils; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-   MA 02110-1301 USA */
+   along with GNU Mailutils.  If not, see <http://www.gnu.org/licenses/>. */
 
 #include <mh.h>
 
@@ -30,12 +29,12 @@ static void
 addrcp (mu_list_t *list, char *addr, int isbcc)
 {
   int rc;
-  struct recipient *p = xmalloc (sizeof (*p));
+  struct recipient *p = mu_alloc (sizeof (*p));
   p->addr = addr;
   p->isbcc = isbcc;
   if (!*list && (rc = mu_list_create (list)))
     {
-      mu_error (_("Cannot create list: %s"), mu_strerror (rc));
+      mu_error (_("cannot create list: %s"), mu_strerror (rc));
       exit (1);
     }
   mu_list_append (*list, p);
@@ -48,72 +47,60 @@ ismydomain (char *p)
   if (!p)
     return 1;
   mu_get_user_email_domain (&domain);
-  return strcasecmp (domain, p + 1) == 0;
+  return mu_c_strcasecmp (domain, p + 1) == 0;
 }
 
+/* FIXME: incl is not used */
 int
 mh_alias_expand (const char *str, mu_address_t *paddr, int *incl)
 {
-  int argc;
-  char **argv;
-  int i;
-  char *buf;
-  mu_address_t exaddr = NULL;
+  size_t i, count;
+  mu_address_t addr;
+  int status;
+  
+  if (!str || !*str)
+    {
+      *paddr = NULL;
+      return 0;
+    }
 
   if (incl)
     *incl = 0;
-  mu_argcv_get (str, ",", NULL, &argc, &argv);
-  for (i = 0; i < argc;)
+  status = mu_address_create_hint (&addr, str, NULL, 0);
+  if (status)
     {
-      if (i + 1 == argc)
-	{
-	  if (mh_alias_get_address (argv[i], &exaddr, incl) == 0)
-	    {
-	      free (argv[i]);
-	      memcpy (&argv[i], &argv[i+1],
-		      (argc - i + 1) * sizeof (argv[0]));
-	      argc--;
-	    }
-	  else
-	    i++;
-	}
-      else if (argv[i + 1][0] == ',')
-	{
-	  if (mh_alias_get_address (argv[i], &exaddr, incl) == 0)
-	    {
-	      free (argv[i]);
-	      free (argv[i+1]);
-	      memcpy (&argv[i], &argv[i+2],
-		      (argc - i) * sizeof (argv[0]));
-	      argc -= 2;
-	    }
-	  else
-	    i += 2;
-	}
-      else
-	{
-	  for (; i < argc; i++)
-	    if (argv[i][0] == ',')
-	      {
-		i++;
-		break;
-	      }
-	}
+      mu_error (_("Bad address `%s': %s"), str, mu_strerror (status));
+      return 1;
     }
 
-  if (argc)
+  mu_address_get_count (addr, &count);
+  for (i = 1; i <= count; i++)
     {
-      int status;
-      mu_argcv_string (argc, argv, &buf);
-      if ((status = mu_address_create (paddr, buf)))
-	mu_error (_("Bad address `%s': %s"), buf, mu_strerror (status));
-      free (buf);
-    }
+      mu_address_t subaddr = NULL;
+      const char *key;
 
-  mu_argcv_free (argc, argv);
-  
-  mu_address_union (paddr, exaddr);
-  mu_address_destroy (&exaddr);
+      if (mu_address_sget_domain (addr, i, &key) == 0 && key == NULL)
+	{
+	  if (mu_address_sget_local_part (addr, i, &key) ||
+	      mh_alias_get_address (key, paddr, incl) == 0 ||
+	      /* Recreating address from local part adds the default
+		 domain to it: */
+              mu_address_create (&subaddr, key))
+            continue;
+	}
+      else 
+        {
+          status = mu_address_get_nth (addr, i, &subaddr);
+          if (status)
+	    {
+	      mu_error (_("%s: cannot get address #%lu: %s"),
+			str, (unsigned long) i, mu_strerror (status));
+	      continue;
+	    }
+        }   
+      mu_address_union (paddr, subaddr);
+      mu_address_destroy (&subaddr);
+    }
   return 0;
 }
 
@@ -169,7 +156,7 @@ destroy_addrs (mu_list_t *list)
 {
   if (!*list)
     return;
-  mu_list_do (*list, _destroy_recipient, NULL);
+  mu_list_foreach (*list, _destroy_recipient, NULL);
   mu_list_destroy (list);
 }
 
@@ -214,51 +201,121 @@ _print_local_recipient (void *item, void *data)
   (*count)++;
   return 0;
 }
-		  
+
+static mu_header_t
+read_header (mu_stream_t stream)
+{
+  int rc;
+  mu_stream_t flt;
+  mu_off_t size;
+  size_t total;
+  char *blurb;
+  mu_header_t hdr;
+  
+  rc = mu_stream_size (stream, &size);
+  if (rc)
+    {
+      mu_error (_("cannot get stream size: %s"), mu_strerror (rc));
+      exit (1);
+    }
+  
+  rc = mu_filter_create (&flt, stream, "HEADER",
+			 MU_FILTER_DECODE, MU_STREAM_READ);
+  if (rc)
+    {
+      mu_error (_("cannot open filter stream: %s"), mu_strerror (rc));
+      exit (1);
+    }
+
+  blurb = mu_alloc (size + 1);
+
+  total = 0;
+  while (1)
+    {
+      size_t n;
+      
+      rc = mu_stream_read (flt, blurb + total, size - total, &n);
+      if (rc)
+	break;
+      if (n == 0)
+	break;
+      total += n;
+    }
+
+  mu_stream_destroy (&flt);
+  if (rc)
+    {
+      free (blurb);
+      mu_error (_("read error: %s"), mu_strerror (rc));
+      exit (1);
+    }
+
+  rc = mu_header_create (&hdr, blurb, total);
+  free (blurb);
+  if (rc)
+    {
+      mu_diag_funcall (MU_DIAG_ERROR, "mu_header_create", NULL, rc);
+      exit (1);
+    }
+  return hdr;
+}
+
 int
 mh_whom (const char *filename, int check)
 {
   int rc = 0;
-  mh_context_t *ctx;
 
-  mh_read_aliases ();
-  ctx = mh_context_create (filename, 1);
-  if ((rc = mh_context_read (ctx)))
+  if (access (filename, R_OK))
     {
-      if (rc == ENOENT)
-	mu_error ("%s: %s", filename, mu_strerror (rc));
-      else
-	mu_error ("%s: %s (%s)", filename, _("Malformed message"),
-		  mu_strerror (rc));
+      mu_error ("%s: %s", filename, mu_strerror (errno));
       rc = -1;
     }
   else
     {
       size_t count = 0;
-      
-      scan_addrs (mh_context_get_value (ctx, MU_HEADER_TO, NULL), 0);
-      scan_addrs (mh_context_get_value (ctx, MU_HEADER_CC, NULL), 0);
-      scan_addrs (mh_context_get_value (ctx, MU_HEADER_BCC, NULL), 1);
+      mu_header_t hdr;
+      mu_stream_t str;
+      int rc;
+      const char *val;
+
+      rc = mu_file_stream_create (&str, filename, MU_STREAM_READ);
+      if (rc)
+	{
+	  mu_diag_funcall (MU_DIAG_ERROR, "mu_file_stream_create",
+			   filename, rc);
+	  exit (1);
+	}
+      hdr = read_header (str);
+      mu_stream_unref (str);
+
+      mh_read_aliases ();
+
+      if (mu_header_sget_value (hdr, MU_HEADER_TO, &val) == 0)
+	scan_addrs (val, 0);
+      if (mu_header_sget_value (hdr, MU_HEADER_CC, &val) == 0)
+	scan_addrs (val, 0);
+      if (mu_header_sget_value (hdr, MU_HEADER_BCC, &val) == 0)
+	scan_addrs (val, 1);
 
       if (local_rcp)
 	{
 	  printf ("  %s\n", _("-- Local Recipients --"));
-	  mu_list_do (local_rcp, _print_local_recipient, &count);
+	  mu_list_foreach (local_rcp, _print_local_recipient, &count);
 	}
 
       if (network_rcp)
 	{
 	  printf ("  %s\n", _("-- Network Recipients --"));
-	  mu_list_do (network_rcp, _print_recipient, &count);
+	  mu_list_foreach (network_rcp, _print_recipient, &count);
 	}
 
       if (count == 0)
 	{
-	  mu_error(_("No recipients"));
+	  mu_error(_("no recipients"));
 	  rc = -1;
 	}
+      mu_header_destroy (&hdr);
     }
-  free (ctx);
   destroy_addrs (&network_rcp);
   destroy_addrs (&local_rcp);
   return rc;
