@@ -1,6 +1,6 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Free
-   Software Foundation, Inc.
+   Copyright (C) 2003, 2005-2012, 2014-2016 Free Software Foundation,
+   Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,43 +27,9 @@
 #include <unistd.h>
 #include <mailutils/mailutils.h>
 #include <mailutils/tls.h>
-#include "mailutils/libargp.h"
+#include "mailutils/cli.h"
 #include <muaux.h>
 
-static char doc[] = N_("GNU movemail -- move messages across mailboxes.");
-static char args_doc[] = N_("inbox-url destfile [POP-password]");
-
-enum {
-  EMACS_OPTION=256,
-  IGNORE_ERRORS_OPTION,
-  PROGRAM_ID_OPTION,
-  MAX_MESSAGES_OPTION,
-  ONERROR_OPTION
-};
-
-static struct argp_option options[] = {
-  { "preserve", 'p', NULL, 0, N_("preserve the source mailbox") },
-  { "keep-messages", 0, NULL, OPTION_ALIAS, NULL },
-  { "reverse",  'r', NULL, 0, N_("reverse the sorting order") },
-  { "emacs", EMACS_OPTION, NULL, 0,
-    N_("output information used by Emacs rmail interface") },
-  { "uidl", 'u', NULL, 0,
-    N_("use UIDLs to avoid downloading the same message twice") },
-  { "verbose", 'v', NULL, 0,
-    N_("increase verbosity level") },
-  { "owner", 'P', N_("MODELIST"), 0,
-    N_("control mailbox ownership") },
-  { "ignore-errors", IGNORE_ERRORS_OPTION, NULL, 0,
-    N_("try to continue after errors") },
-  { "onerror", ONERROR_OPTION, N_("KW[,KW...]"), 0,
-    N_("what to do on errors") },
-  { "program-id", PROGRAM_ID_OPTION, N_("FMT"), 0,
-    N_("set program identifier for diagnostics (default: program name)") },
-  { "max-messages", MAX_MESSAGES_OPTION, N_("NUMBER"), 0,
-    N_("process at most NUMBER messages") },
-  { NULL,      0, NULL, 0, NULL, 0 }
-};
-
 static int reverse_order;
 static int preserve_mail; 
 static int emacs_mode;
@@ -72,6 +38,7 @@ static int verbose_option;
 static int ignore_errors;
 static char *program_id_option;
 static size_t max_messages_option;
+static int notify;
 
   /* These bits tell what to do when an error occurs: */
 #define ONERROR_SKIP     0x01  /* Skip to the next message */
@@ -131,80 +98,9 @@ mu_kwd_t method_kwd[] = {
   { "set-id", set_owner_id },
   { NULL }
 };
-
-static error_t
-parse_opt (int key, char *arg, struct argp_state *state)
-{
-  static mu_list_t lst;
-
-  switch (key)
-    {
-    case 'r':
-      mu_argp_node_list_new (lst, "reverse", "yes");
-      break;
-
-    case 'p':
-      mu_argp_node_list_new (lst, "preserve", "yes");
-      break;
-
-    case 'P':
-      mu_argp_node_list_new (lst, "mailbox-ownership", arg);
-      break;
-
-    case 'u':
-      mu_argp_node_list_new (lst, "uidl", "yes");
-      break;
-
-    case 'v':
-      verbose_option++;
-      break;
-      
-    case EMACS_OPTION:
-      mu_argp_node_list_new (lst, "emacs", "yes");
-      break;
-
-    case IGNORE_ERRORS_OPTION:
-      mu_argp_node_list_new (lst, "ignore-errors", "yes");
-      break;
-
-    case ONERROR_OPTION:
-      mu_argp_node_list_new (lst, "onerror", arg);
-      break;
-      
-    case MAX_MESSAGES_OPTION:
-      mu_argp_node_list_new (lst, "max-messages", arg);
-      break;
-      
-    case PROGRAM_ID_OPTION:
-      mu_argp_node_list_new (lst, "program-id", arg);
-      break;
-      
-    case ARGP_KEY_INIT:
-      mu_argp_node_list_init (&lst);
-      break;
-      
-    case ARGP_KEY_FINI:
-      mu_argp_node_list_finish (lst, NULL, NULL);
-      break;
-      
-    default:
-      return ARGP_ERR_UNKNOWN;
-    }
-  return 0;
-}
-
-static struct argp argp = {
-  options,
-  parse_opt,
-  args_doc,
-  doc,
-  NULL,
-  NULL, NULL
-};
-
 
 static int
-_cb_mailbox_ownership (const char *str)
+set_mailbox_ownership (const char *str)
 {
   if (strcmp (str, "clear") == 0)
     so_method_num = 0;
@@ -271,35 +167,152 @@ _cb_mailbox_ownership (const char *str)
   return 0;
 }
 
+static int
+set_mailbox_ownership_list (char const *str)
+{
+  if (!strchr (str, ','))
+    return set_mailbox_ownership (str);
+  else
+    {
+      struct mu_wordsplit ws;
+      size_t i;
+      
+      ws.ws_delim = ",";
+      if (mu_wordsplit (str, &ws, MU_WRDSF_DEFFLAGS|MU_WRDSF_DELIM))
+	{
+	  mu_error (_("cannot parse %s: %s"),
+		    str, mu_wordsplit_strerror (&ws));
+	  return 1;
+	}
+
+      for (i = 0; i < ws.ws_wordc; i++)
+	if (set_mailbox_ownership (ws.ws_wordv[i]))
+	  return 1;
+      mu_wordsplit_free (&ws);
+      return 0;
+    }
+}
+
+static int
+set_onerror_action (void *item, void *data)
+{
+  char *str = item;
+
+  if (strcmp (str, "abort") == 0)
+    onerror_flags = 0;
+  else
+    {
+      static struct mu_kwd onerror_kw[] = {
+	{ "skip", ONERROR_SKIP },
+	{ "delete", ONERROR_DELETE },
+	{ "count", ONERROR_COUNT },
+	{ NULL }
+      };
+      int flag, clr = 0;
+
+      if (strncmp (str, "no", 2) == 0)
+	{
+	  clr = 1;
+	  str += 2;
+	}
+      if (mu_kwd_xlat_name (onerror_kw, str, &flag))
+	{
+	  mu_error (_("unknown keyword: %s"), str);
+	  return 1;
+	}
+      if (clr)
+	onerror_flags &= ~flag;
+      else
+	onerror_flags |= flag;
+    }
+  return 0;
+}
+
+static int
+set_onerror_actions (char const *str)
+{
+  mu_list_t list;
+  int rc;
+  
+  mu_list_create (&list);
+  mu_list_set_destroy_item (list, mu_list_free_item);
+  mu_string_split (str, ",", list);
+  rc = mu_list_foreach (list, set_onerror_action, NULL);
+  mu_list_destroy (&list);
+  return rc;
+}
+
+static void
+cli_mailbox_ownership (struct mu_parseopt *po, struct mu_option *opt,
+		       char const *arg)
+{
+  if (set_mailbox_ownership_list (arg))
+    exit (po->po_exit_error);
+}
+
+static void
+cli_onerror (struct mu_parseopt *po, struct mu_option *opt, char const *arg)
+{
+  if (set_onerror_actions (arg))
+    exit (po->po_exit_error);
+}
+
+static struct mu_option movemail_options[] = {
+  { "preserve", 'p', NULL, MU_OPTION_DEFAULT,
+    N_("preserve the source mailbox"),
+    mu_c_bool, &preserve_mail },
+  { "keep-messages", 0, NULL, MU_OPTION_ALIAS },
+  
+  { "reverse",  'r', NULL, MU_OPTION_DEFAULT,
+    N_("reverse the sorting order"),
+    mu_c_bool, &reverse_order },
+  
+  { "emacs",      0, NULL, MU_OPTION_DEFAULT,
+    N_("output information used by Emacs rmail interface"),
+    mu_c_bool, &emacs_mode },
+  
+  { "uidl",     'u', NULL, MU_OPTION_DEFAULT,
+    N_("use UIDLs to avoid downloading the same message twice"),
+    mu_c_bool, &uidl_option },
+  
+  { "verbose",  'v', NULL, MU_OPTION_DEFAULT,
+    N_("increase verbosity level"),
+    mu_c_incr, &verbose_option },
+  
+  { "owner",    'P', N_("MODELIST"), MU_OPTION_DEFAULT,
+    N_("control mailbox ownership"),
+    mu_c_string, cli_mailbox_ownership },
+  
+  { "ignore-errors", 0, NULL, MU_OPTION_DEFAULT,
+    N_("try to continue after errors"),
+    mu_c_bool, &ignore_errors },
+  
+  { "onerror",       0, N_("KW[,KW...]"), MU_OPTION_DEFAULT,
+    N_("what to do on errors"),
+    mu_c_string, NULL, cli_onerror },
+  
+  { "program-id",    0, N_("FMT"), MU_OPTION_DEFAULT,
+    N_("set program identifier for diagnostics (default: program name)"),
+    mu_c_string, &program_id_option },
+  
+  { "max-messages",  0, N_("NUMBER"), MU_OPTION_DEFAULT,
+    N_("process at most NUMBER messages"),
+    mu_c_size, &max_messages_option },
+  
+  { "notify",        0, NULL,   MU_OPTION_DEFAULT,
+    N_("enable biff notification"),
+    mu_c_bool, &notify },
+
+  MU_OPTION_END
+}, *options[] = { movemail_options, NULL };
+
 static int
 cb_mailbox_ownership (void *data, mu_config_value_t *val)
 {
   int i;
   
   if (val->type == MU_CFG_STRING)
-    {
-      const char *str = val->v.string;
-      if (!strchr (str, ','))
-	return _cb_mailbox_ownership (str);
-      else
-	{
-	  struct mu_wordsplit ws;
-
-	  ws.ws_delim = ",";
-	  if (mu_wordsplit (str, &ws, MU_WRDSF_DEFFLAGS|MU_WRDSF_DELIM))
-	    {
-	      mu_error (_("cannot parse %s: %s"),
-			str, mu_wordsplit_strerror (&ws));
-	      return 1;
-	    }
-
-	  for (i = 0; i < ws.ws_wordc; i++)
-	    if (_cb_mailbox_ownership (ws.ws_wordv[i]))
-	      return 1;
-	  mu_wordsplit_free (&ws);
-	  return 0;
-	}
-    }
+    set_mailbox_ownership_list (val->v.string);
 		
   if (mu_cfg_assert_value_type (val, MU_CFG_LIST))
     return 1;
@@ -308,7 +321,7 @@ cb_mailbox_ownership (void *data, mu_config_value_t *val)
     {
       if (mu_cfg_assert_value_type (&val->v.arg.v[i], MU_CFG_STRING))
 	return 1;
-      if (_cb_mailbox_ownership (val->v.arg.v[i].v.string))
+      if (set_mailbox_ownership (val->v.arg.v[i].v.string))
 	return 1;
     }
   return 0;
@@ -317,63 +330,34 @@ cb_mailbox_ownership (void *data, mu_config_value_t *val)
 static int
 cb_onerror (void *data, mu_config_value_t *val)
 {
-  struct mu_wordsplit ws;
-  static struct mu_kwd onerror_kw[] = {
-    { "skip", ONERROR_SKIP },
-    { "delete", ONERROR_DELETE },
-    { "count", ONERROR_COUNT },
-    { NULL }
-  };
-  int i, flag;
-  
-  if (mu_cfg_assert_value_type (val, MU_CFG_STRING))
-    return 1;
-  if (strcmp (val->v.string, "abort") == 0)
+  switch (val->type)
     {
-      onerror_flags = 0;
-      return 0;
+    case MU_CFG_LIST:
+      mu_list_foreach (val->v.list, set_onerror_action, NULL);
+      break;
+
+    case MU_CFG_STRING:
+      set_onerror_actions (val->v.string);
+      break;
+
+    default:
+      mu_error ("%s", _("too many arguments"));
     }
-  ws.ws_delim = ",";
-  if (mu_wordsplit (val->v.string, &ws,
-		    MU_WRDSF_NOVAR | MU_WRDSF_NOCMD |
-		    MU_WRDSF_DELIM | MU_WRDSF_WS))
-    {
-      mu_error (_("cannot split argument: %s"), mu_wordsplit_strerror (&ws));
-      return 1;
-    }
-  for (i = 0; i < ws.ws_wordc; i++)
-    {
-      int clr = 0;
-      char *name = ws.ws_wordv[i];
-      
-      if (strncmp (name, "no", 2) == 0)
-	{
-	  clr = 1;
-	  name += 2;
-	}
-      if (mu_kwd_xlat_name (onerror_kw, name, &flag))
-	mu_error (_("unknown keyword: %s"), ws.ws_wordv[i]);
-      if (clr)
-	onerror_flags &= ~flag;
-      else
-	onerror_flags |= flag;
-    }
-  mu_wordsplit_free (&ws);
   return 0;
 }
   
 struct mu_cfg_param movemail_cfg_param[] = {
-  { "preserve", mu_cfg_bool, &preserve_mail, 0, NULL,
+  { "preserve", mu_c_bool, &preserve_mail, 0, NULL,
     N_("Do not remove messages from the source mailbox.") },
-  { "reverse",  mu_cfg_bool, &reverse_order, 0, NULL,
+  { "reverse",  mu_c_bool, &reverse_order, 0, NULL,
     N_("Reverse message sorting order.") },
-  { "emacs", mu_cfg_bool, &emacs_mode, 0, NULL,
+  { "emacs", mu_c_bool, &emacs_mode, 0, NULL,
     N_("Output information used by Emacs rmail interface.") },
-  { "uidl", mu_cfg_bool, &uidl_option, 0, NULL,
+  { "uidl", mu_c_bool, &uidl_option, 0, NULL,
     N_("Use UIDLs to avoid downloading the same message twice.") },
-  { "verbose", mu_cfg_int, &verbose_option, 0, NULL,
+  { "verbose", mu_c_int, &verbose_option, 0, NULL,
     N_("Set verbosity level.") },
-  { "program-id", mu_cfg_string, &program_id_option, 0, NULL,
+  { "program-id", mu_c_string, &program_id_option, 0, NULL,
     N_("Set program identifier string (default: program name)") },
   { "mailbox-ownership", mu_cfg_callback, NULL, 0,
     cb_mailbox_ownership,
@@ -384,24 +368,30 @@ struct mu_cfg_param movemail_cfg_param[] = {
        " set-id=UID[:GID] set supplied UID and GID\n"
        " set-name=USER    make destination mailbox owned by USER"),
     N_("methods: list") },
-  { "max-messages", mu_cfg_size, &max_messages_option, 0, NULL,
+  { "max-messages", mu_c_size, &max_messages_option, 0, NULL,
     N_("Copy at most <count> messages."),
     N_("count") },
-  { "ignore-errors", mu_cfg_bool, &ignore_errors, 0, NULL,
+  { "ignore-errors", mu_c_bool, &ignore_errors, 0, NULL,
     N_("Continue after an error.") },
   { "onerror", mu_cfg_callback, NULL, 0, cb_onerror,
-    N_("What to do after an error. Argument is a comma-separated list of:\n"
+    N_("What to do after an error. Argument is a list of:\n"
+       " abort  -  terminate the program (the default)\n"
        " skip   -  skip to the next message\n"
        " delete -  delete this one and to the next message\n"
        " count  -  count this message as processed\n"
-       "Each keyword can be prefixed with \"no\" to reverse its meaning\n"
-       "Setting onerror=abort reverts to the default behavior.") },
+       "Each keyword can be prefixed with \"no\" to reverse its meaning."),
+    N_("arg: list") },
   { NULL }
 };
-
 
-static const char *movemail_capa[] = {
-  "common",
+struct mu_cli_setup cli = {
+  options,
+  movemail_cfg_param,
+  N_("GNU movemail -- move messages across mailboxes."),
+  N_("inbox-url destfile [POP-password]")
+};
+
+static char *movemail_capa[] = {
   "debug",
   "locking",
   "mailbox",
@@ -748,46 +738,66 @@ struct movemail_getvar_closure
 #define SEQ(s, n, l) \
   (((l) == (sizeof(s) - 1)) && memcmp (s, n, l) == 0)
 
-static const char *
-get_url_part (mu_url_t url, const char *name, size_t nlen)
+static int
+get_url_part (mu_url_t url, const char *name, size_t nlen, char **ret)
 {
-  int rc = MU_ERR_NOENT;
-  const char *s;
+  int rc;
   
   if (!url)
-    return NULL;
+    return MU_WRDSE_UNDEF;
   if (SEQ ("user", name, nlen))
-    rc = mu_url_sget_user (url, &s);
+    rc = mu_url_aget_user (url, ret);
   else if (SEQ ("host", name, nlen))
-    rc = mu_url_sget_host (url, &s);
+    rc = mu_url_aget_host (url, ret);
   else if (SEQ ("port", name, nlen))
-    rc = mu_url_sget_portstr (url, &s);
+    rc = mu_url_aget_portstr (url, ret);
   else if (SEQ ("path", name, nlen))
-    rc = mu_url_sget_path (url, &s);
+    rc = mu_url_aget_path (url, ret);
+  else
+    return MU_WRDSE_UNDEF;
 
-  if (rc)
-    return NULL;
-  return s;
+  switch (rc)
+    {
+    case 0:
+      break;
+      
+    case MU_ERR_NOENT:
+      return MU_WRDSE_UNDEF;
+
+    default:
+      if (mu_asprintf (ret, "%s", mu_strerror (rc)))
+	return MU_WRDSE_NOSPACE;
+      return MU_WRDSE_USERERR;
+    }
+  
+  return MU_WRDSE_OK;
 }
     
-static const char *
-movemail_getvar (const char *name, size_t nlen, void *data)
+static int
+movemail_getvar (char **ret, const char *name, size_t nlen, void *data)
 {
   struct movemail_getvar_closure *pc = data;
-  
-  if (SEQ ("progname", name, nlen))
-    return mu_program_name;
-  if (SEQ ("source", name, nlen))
-    return pc->source_name;
-  if (SEQ ("dest", name, nlen))
-    return pc->dest_name;
+  const char *s;
 
   if (nlen > 7 && memcmp ("source_", name, 7) == 0)
-    return get_url_part (pc->source_url, name + 7, nlen - 7);
+    return get_url_part (pc->source_url, name + 7, nlen - 7, ret);
+
   if (nlen > 5 && memcmp ("dest_", name, 5) == 0)
-    return get_url_part (pc->dest_url, name + 5, nlen - 5);
+    return get_url_part (pc->dest_url, name + 5, nlen - 5, ret);
   
-  return NULL;
+  if (SEQ ("progname", name, nlen))
+    s = mu_program_name;
+  else if (SEQ ("source", name, nlen))
+    s = pc->source_name;
+  else if (SEQ ("dest", name, nlen))
+    s = pc->dest_name;
+  else
+    return MU_WRDSE_UNDEF;
+    
+  *ret = strdup (s);
+  if (!*ret)
+    return MU_WRDSE_NOSPACE;
+  return MU_WRDSE_OK;
 }
 
 static void
@@ -825,7 +835,6 @@ set_program_id (const char *source_name, const char *dest_name)
      plays wise with its argument. We need a mu_set_diag_prefix
      function. */
   mu_program_name = ws.ws_wordv[0];
-  ws.ws_wordv[0] = NULL;
   ws.ws_wordc = 0;
   mu_wordsplit_free (&ws);
   mu_stdstream_strerr_setup (MU_STRERR_STDERR);
@@ -835,7 +844,6 @@ set_program_id (const char *source_name, const char *dest_name)
 int
 main (int argc, char **argv)
 {
-  int index;
   size_t total;
   int rc = 0;
   char *source_name, *dest_name;
@@ -849,19 +857,11 @@ main (int argc, char **argv)
   
   /* Register the desired "mailbox" formats.  */
   mu_register_all_formats ();
-
-  /* argument parsing */
+  /* Register authentication modules */
+  mu_auth_register_module (&mu_auth_tls_module);
   
-#ifdef WITH_TLS
-  mu_gocs_register ("tls", mu_tls_module_init);
-#endif
-  mu_argp_init (NULL, NULL);
-  if (mu_app_init (&argp, movemail_capa, movemail_cfg_param, 
-		   argc, argv, 0, &index, NULL))
-    exit (1);
-
-  argc -= index;
-  argv += index;
+  /* argument parsing */
+  mu_cli (argc, argv, &cli, movemail_capa, NULL, &argc, &argv);
 
   if (argc < 2 || argc > 3)
     {
@@ -894,11 +894,20 @@ main (int argc, char **argv)
 
   switch_owner (source);
   
-  open_mailbox (&dest, dest_name, MU_STREAM_RDWR | MU_STREAM_CREAT, NULL);
+  open_mailbox (&dest, dest_name,
+		MU_STREAM_APPEND | MU_STREAM_READ | MU_STREAM_CREAT, NULL);
 
   if (program_id_option)
     set_program_id (source_name, dest_name);
 
+  if (notify)
+    {
+      rc = mu_mailbox_set_notify (dest, NULL);
+      if (rc)
+	mu_error (_("failed to set up notification: %s"),
+		  mu_strerror (rc));
+    }
+  
   rc = mu_mailbox_messages_count (source, &total);
   if (rc)
     {
