@@ -1,6 +1,5 @@
-/*
-   GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009 Free Software Foundation, Inc.
+/* GNU Mailutils -- a suite of utilities for electronic mail
+   Copyright (C) 2009-2012, 2014-2016 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,10 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA
-*/
+   Public License along with this library.  If not, see
+   <http://www.gnu.org/licenses/>. */
 
 #include "libmu_py.h"
 
@@ -90,7 +87,7 @@ api_header_size (PyObject *self, PyObject *args)
     return NULL;
 
   status = mu_header_size (py_hdr->hdr, &size);
-  return status_object (status, PyInt_FromLong (size));
+  return status_object (status, PyInt_FromSize_t (size));
 }
 
 static PyObject *
@@ -104,7 +101,7 @@ api_header_lines (PyObject *self, PyObject *args)
     return NULL;
 
   status = mu_header_lines (py_hdr->hdr, &lines);
-  return status_object (status, PyInt_FromLong (lines));
+  return status_object (status, PyInt_FromSize_t (lines));
 }
 
 static PyObject *
@@ -125,14 +122,15 @@ api_header_get_value (PyObject *self, PyObject *args)
 static PyObject *
 api_header_get_value_n (PyObject *self, PyObject *args)
 {
-  int status, n;
+  int status;
+  Py_ssize_t n;
   char *name;
   const char *value = NULL;
   PyHeader *py_hdr;
 
-  if (!PyArg_ParseTuple (args, "O!si", &PyHeaderType, &py_hdr, &name, &n))
+  if (!PyArg_ParseTuple (args, "O!sn", &PyHeaderType, &py_hdr, &name, &n))
     return NULL;
-
+  ASSERT_INDEX_RANGE (n, "header");
   status = mu_header_sget_value_n (py_hdr->hdr, name, n, &value);
   return status_object (status, PyString_FromString (value ? value : ""));
 }
@@ -163,20 +161,20 @@ api_header_get_field_count (PyObject *self, PyObject *args)
     return NULL;
 
   status = mu_header_get_field_count (py_hdr->hdr, &count);
-  return status_object (status, PyInt_FromLong (count));
+  return status_object (status, PyInt_FromSize_t (count));
 }
 
 static PyObject *
 api_header_get_field_name (PyObject *self, PyObject *args)
 {
   int status;
-  size_t idx;
+  Py_ssize_t idx;
   const char *name = NULL;
   PyHeader *py_hdr;
 
-  if (!PyArg_ParseTuple (args, "O!i", &PyHeaderType, &py_hdr, &idx))
+  if (!PyArg_ParseTuple (args, "O!n", &PyHeaderType, &py_hdr, &idx))
     return NULL;
-
+  ASSERT_INDEX_RANGE (idx, "header");
   status = mu_header_sget_field_name (py_hdr->hdr, idx, &name);
   return status_object (status, PyString_FromString (name ? name : ""));
 }
@@ -185,13 +183,13 @@ static PyObject *
 api_header_get_field_value (PyObject *self, PyObject *args)
 {
   int status;
-  size_t idx;
+  Py_ssize_t idx;
   const char *value = NULL;
   PyHeader *py_hdr;
 
-  if (!PyArg_ParseTuple (args, "O!i", &PyHeaderType, &py_hdr, &idx))
+  if (!PyArg_ParseTuple (args, "O!n", &PyHeaderType, &py_hdr, &idx))
     return NULL;
-
+  ASSERT_INDEX_RANGE (idx, "header");
   status = mu_header_sget_field_value (py_hdr->hdr, idx, &value);
   return status_object (status, PyString_FromString (value ? value : ""));
 }
